"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.activate = void 0;
const vscode = require("vscode");
const pkg = require('../package.json');
const disposable_1 = require("./disposable");
const foldingProvider_1 = require("./foldingProvider");
const DEPRECATED_KEY = 'explicitFoldingDeprecated';
const VERSION_KEY = 'explicitFoldingVersion';
const SCHEMES = ['file', 'untitled', 'vscode-userdata'];
let $channel = null;
let $context = null;
const $disposable = new disposable_1.Disposable();
class MainProvider {
    constructor() {
        this.providers = {};
        this.id = 'explicit';
    }
    provideFoldingRanges(document) {
        if (!this.providers[document.languageId]) {
            this.providers[document.languageId] = true;
            const config = vscode.workspace.getConfiguration('explicitFolding', document);
            const delay = getDelay(config);
            if (delay > 0) {
                setTimeout(() => this.setup(document), delay);
            }
            else {
                this.setup(document);
            }
        }
        return [];
    } // }}}
    setup(document) {
        const language = document.languageId;
        const perLanguage = getRules(vscode.workspace.getConfiguration('explicitFolding'));
        const config = vscode.workspace.getConfiguration('explicitFolding', document);
        const rules = [];
        if (!applyRules(config.get('rules'), rules)) {
            applyRules(perLanguage[language], rules);
        }
        applyRules(perLanguage['*'], rules);
        checkDeprecatedRules(rules);
        const debug = config.get('debug') || false;
        const channel = getDebugChannel(debug);
        const provider = new foldingProvider_1.default(rules, channel);
        for (const scheme of SCHEMES) {
            const disposable = vscode.languages.registerFoldingRangeProvider({ language, scheme }, provider);
            $disposable.push(disposable);
        }
    } // }}}
}
function applyRules(data, rules) {
    if (data) {
        if (Array.isArray(data)) {
            rules.push(...data);
        }
        else {
            rules.push(data);
        }
        return true;
    }
    else {
        return false;
    }
} // }}}
function checkDeprecatedRule(rule, deprecateds) {
    if (rule instanceof Array) {
        for (const r of rule) {
            checkDeprecatedRule(r, deprecateds);
        }
    }
    else {
        if (rule.descendants) {
            if (!deprecateds.includes('descendants')) {
                deprecateds.push('descendants');
            }
        }
        else if (Array.isArray(rule.nested)) {
            for (const r of rule.nested) {
                checkDeprecatedRule(r, deprecateds);
            }
        }
    }
} // }}}
function checkDeprecatedRules(rules) {
    const deprecateds = [];
    checkDeprecatedRule(rules, deprecateds);
    if (deprecateds.includes('descendants')) {
        vscode.window.showWarningMessage('Please update your config. The property `descendants` has been deprecated and replaced with the property `nested`. It will be removed in the next version.');
    }
} // }}}
function getDelay(config) {
    if (config.has('startupDelay')) {
        vscode.window.showWarningMessage('Please update your config. The property `startupDelay` has been deprecated and replaced with the property `delay`. It will be removed in the next version.');
        return config.get('startupDelay') || 0;
    }
    else {
        return config.get('delay') || 0;
    }
} // }}}
function getRules(config) {
    let rules = vscode.workspace.getConfiguration('folding');
    if (rules) {
        const value = $context.globalState.get(DEPRECATED_KEY);
        const lastWarning = value ? new Date(value) : null;
        const currentWarning = new Date();
        if (currentWarning > new Date(2022, 6, 1)) {
            vscode.window.showErrorMessage('Please update your config. The property `folding` is not supported since July 1, 2022. It has been replaced with the property `explicitFolding.rules`.');
            return config.rules;
        }
        else if (!lastWarning || lastWarning.getFullYear() !== currentWarning.getFullYear() || lastWarning.getMonth() !== currentWarning.getMonth() || currentWarning > new Date(2022, 5, 1)) {
            $context.globalState.update(DEPRECATED_KEY, currentWarning);
            vscode.window.showWarningMessage('Please update your config. The property `folding` has been deprecated and replaced with the property `explicitFolding.rules`. Its support will stop on July 1, 2022.');
        }
        return rules;
    }
    else {
        return config.rules;
    }
} // }}}
function getDebugChannel(debug) {
    if (debug) {
        if (!$channel) {
            $channel = vscode.window.createOutputChannel('Folding');
        }
        return $channel;
    }
    else {
        return null;
    }
} // }}}
function setup() {
    $disposable.dispose();
    const provider = new MainProvider();
    for (const scheme of SCHEMES) {
        const disposable = vscode.languages.registerFoldingRangeProvider({ language: '*', scheme }, provider);
        $disposable.push(disposable);
    }
    $context.subscriptions.push($disposable);
} // }}}
function showWhatsNewMessage(version) {
    return __awaiter(this, void 0, void 0, function* () {
        const actions = [{
                title: 'Homepage'
            }, {
                title: 'Release Notes'
            }];
        const result = yield vscode.window.showInformationMessage(`Explicit Folding has been updated to v${version} — check out what's new!`, ...actions);
        if (result != null) {
            if (result === actions[0]) {
                yield vscode.commands.executeCommand('vscode.open', vscode.Uri.parse(`${pkg.homepage}`));
            }
            else if (result === actions[1]) {
                yield vscode.commands.executeCommand('vscode.open', vscode.Uri.parse(`${pkg.homepage}/blob/master/CHANGELOG.md`));
            }
        }
    });
} // }}}
function activate(context) {
    return __awaiter(this, void 0, void 0, function* () {
        $context = context;
        const previousVersion = context.globalState.get(VERSION_KEY);
        const currentVersion = pkg.version;
        const config = vscode.workspace.getConfiguration('explicitFolding');
        if (previousVersion === undefined || currentVersion !== previousVersion) {
            context.globalState.update(VERSION_KEY, currentVersion);
            const notification = config.get('notification');
            if (previousVersion === undefined) {
                // don't show notification on install
            }
            else if (notification === 'major') {
                if (currentVersion.split('.')[0] > previousVersion.split('.')[0]) {
                    showWhatsNewMessage(currentVersion);
                }
            }
            else if (notification === 'minor') {
                if (currentVersion.split('.')[0] > previousVersion.split('.')[0] || (currentVersion.split('.')[0] === previousVersion.split('.')[0]) && currentVersion.split('.')[1] > previousVersion.split('.')[1]) {
                    showWhatsNewMessage(currentVersion);
                }
            }
            else if (notification !== 'none') {
                showWhatsNewMessage(currentVersion);
            }
        }
        setup();
        vscode.workspace.onDidChangeConfiguration(event => {
            if (event.affectsConfiguration('folding') || event.affectsConfiguration('explicitFolding')) {
                setup();
            }
        });
    });
} // }}}
exports.activate = activate;
